####################################################
#
# This Work is written by Nikolai Rozanov <nikolai>
#
# Contact:  nikolai.rozanov@gmail.com
#
# Copyright (C) 2018-Present Nikolai Rozanov
#
####################################################

####################################################
# IMPORT STATEMENTS
####################################################

# >>>>>>  Native Imports  <<<<<<<

# >>>>>>  Package Imports <<<<<<<
import numpy as np

# >>>>>>  Local Imports   <<<<<<<


####################################################
# CODE
####################################################

def one_hot(listOfIndexes, numClasses):
    """
    returns [[0,0,1,...],...]
    """
    idxes = np.array( listOfIndexes )

    temp = np.eye( numClasses )[ idxes.reshape(-1) ]

    return np.copy(temp)

def transform_to_one_hot(listOfIndexes, numClasses):
    """
    transforms a listOfIndexes into oneHot
    """
    temp = np.squeeze( one_hot(listOfIndexes, numClasses) )

    return np.copy(temp)



def transform_to_valued_one_hot(listOfValues, listOfIndexes, numClasses):
    """
    transforms list of values and indexes to valued one-hot
    """
    vals = np.array(listOfValues)

    oneHots = one_hot(listOfIndexes, numClasses)

    vOneHots = oneHots * vals.reshape(-1,1)

    temp = np.squeeze( vOneHots )

    return np.copy(temp)





####################################################
# MAIN
####################################################


# EOF



####################################################
#
# This Work is written by Nikolai Rozanov <nikolai>
#
# Contact:  nikolai.rozanov@gmail.com
#
# Copyright (C) 2018-Present Nikolai Rozanov
#
####################################################

####################################################
# IMPORT STATEMENTS
####################################################

# >>>>>>  Native Imports  <<<<<<<

# >>>>>>  Package Imports <<<<<<<
import numpy as np

# >>>>>>  Local Imports   <<<<<<<

####################################################
# CODE
####################################################

def average_simulation_runs(agent, environment, simulationFunction, T=20, N=10):
    """
    Averages the simulations:

    -simulation function needs to accept: agent, environment, **params
    -simulation function needs to return: totalReward, arList, NumOfSimulations
    """
    totalRewardList = []
    timeSteps = range(T)
    accumulatedTRList = []

    for _ in range(N):

        totalR, arList, _ = simulationFunction(agent, environment, N=T)

        totalRewardList.append( totalR )

        rewards = [x for _,x in arList]

        accumulatedTRList.extend( list( zip( timeSteps, rewards ) ) )


    totalRewardMu = np.mean(totalRewardList)
    totalRewardStd = np.std(totalRewardList)


    return (totalRewardMu,totalRewardStd), arList, T, accumulatedTRList






####################################################
# MAIN
####################################################


# EOF



####################################################
#
# This Work is written by Nikolai Rozanov <nikolai>
#
# Contact:  nikolai.rozanov@gmail.com
#
# Copyright (C) 2018-Present Nikolai Rozanov
#
####################################################

####################################################
# IMPORT STATEMENTS
####################################################

# >>>>>>  Native Imports  <<<<<<<

# >>>>>>  Package Imports <<<<<<<
import numpy as np

# >>>>>>  Local Imports   <<<<<<<
from barl import estimators
from barl.utils import utils

from .interfaces import BaseStateLessAgent


####################################################
# CODE
####################################################


class StatelessEpsilonQLearning(BaseStateLessAgent):
    """
    standard Q-learning algorithm
    """

    def __init__(self, numActions, epsilon=0.05, epsilonDecayRate=0.99, estimator=None):
        super().__init__(numActions)
        self.epsilon = epsilon
        self.epsilon_decay_rate = epsilonDecayRate
        self.Q = estimators.EmpiricalMean( [numActions] )


    def learn(self, actionRewardTupleList):
        """
        Learns from a list of ActionReward Tuples
        """

        actions, rewards = utils.unzip( actionRewardTupleList )

        self.Q.update( newValueOrList=rewards, indexList=actions )

        # TODO Add multiple decays
        self.decay_epsilon()

    def take_action(self, epsilon=None ):
        """
        takes an action according to epsilon greedy policy
        """
        if type(epsilon)==type(None):
            epsilon = self.epsilon

        tempRand = np.squeeze( np.random.uniform(0,1,1) )

        if tempRand<epsilon:
            action = np.random.choice(self.numActions)

        else:

            action = np.argmax(self.Q.get_estimate())

        return np.squeeze(action)




    def decay_epsilon(self):
        """
        decays epsilon at one time step
        """
        self.epsilon = self.epsilon * self.epsilon_decay_rate





class StatelessThompsonSampling(BaseStateLessAgent):
    """
    standard Q-learning algorithm
    """

    def __init__(self, numActions, epsilon=0.05, epsilonDecayRate=0.99, estimator=None):
        super().__init__(numActions)
        self.epsilon = epsilon
        self.epsilon_decay_rate = epsilonDecayRate
        self.Q = estimators.NormalMean( [numActions] )


    def learn(self, actionRewardTupleList):
        """
        Learns from a list of ActionReward Tuples
        """

        actions, rewards = utils.unzip( actionRewardTupleList )

        self.Q.update( newValueOrList=rewards, indexList=actions )


    def take_action(self, epsilon=None ):
        """
        takes an action according to epsilon greedy policy
        """

        action = np.argmax( self.Q.sample() )

        return np.squeeze(action)





####################################################
# MAIN
####################################################






# EOF



####################################################
#
# This Work is written by Nikolai Rozanov <nikolai>
#
# Contact:  nikolai.rozanov@gmail.com
#
# Copyright (C) 2018-Present Nikolai Rozanov
#
####################################################

####################################################
# IMPORT STATEMENTS
####################################################

# >>>>>>  Native Imports  <<<<<<<

# >>>>>>  Package Imports <<<<<<<
import numpy as np

# >>>>>>  Local Imports   <<<<<<<
from barl.utils import numerical

from .interfaces import BaseEstimator
####################################################
# CODE
####################################################

class EmpiricalMean(BaseEstimator):
    """
    Estimating 1D Vectors
    """
    def __init__(self, size):
        assert len(size) < 2, "EmpiricalMean only does 1D estimates at the moment"
        self.__estimate = np.zeros(size)
        self.__value = np.zeros(size)
        self.__count = np.zeros(size)

        self.__shape = self.__value.shape
        self.__ndim = self.__value.ndim

    def __str__(self):
        """
        Represents the empricial mean as a string
        """
        outStr = "Means: " + str(self.__estimate)
        return outStr

    def update( self , newValueOrList, indexList=None):
        """
        updates the mean
        Can update from ListOf Values and List of Indexes

        """
        newValue = self.__deal_with_valueList( newValueOrList, indexList )
        oneHotIndexList = self.__deal_with_index( indexList )


        if newValue.shape == self.__shape:
            self.__update_single_value(newValue, oneHotIndexList)


        elif (newValue.ndim == self.__ndim+1) and (newValue.shape[1]==self.__shape[0]):
            self.__update_multiple_values(newValue, oneHotIndexList)

        self.__estimate = np.divide(self.__value, self.__count, where=self.__count>0)



    def get_estimate(self):
        """
        returns the current estimate (of the mean)
        """
        return self.__estimate



    def __update_single_value(self, newValue, oneHotIndexList=None):
        """
        updates a single value
        """
        if type(oneHotIndexList)==type(None):
            where = np.ones(self.__shape)
        else:
            where = oneHotIndexList

        # self.__value = np.add( self.__value, newValue, where=where>0 )
        self.__value += newValue

        self.__count += where




    def __update_multiple_values(self, newValue, oneHotIndexList=None):
        """
        updates multiple values
        """
        if type(oneHotIndexList) == type(None):
            self.__count += np.ones(self.__shape) * newValue.shape[0]

        else:
            self.__count += np.sum( oneHotIndexList, axis=0 )

        self.__value += np.sum( newValue, axis=0 )




    def __deal_with_valueList(self, values, index=None):
        """
        creates valued one hot or does nothing
        """

        if type(index) == type(None):
            outValues = np.array(values)
        else:
            outValues = numerical.transform_to_valued_one_hot( listOfValues=values, listOfIndexes=index, numClasses=self.__shape[0] )

        return outValues




    def __deal_with_index(self,index=None):
        """
        creates a one shot or returns None
        """
        if type(index) == type(None):
            return None
        else:
            outIndex = numerical.transform_to_one_hot(index, numClasses=self.__shape[0])
            return outIndex








####################################################
# MAIN
####################################################


# EOF



####################################################
#
# This Work is written by Nikolai Rozanov <nikolai>
#
# Contact:  nikolai.rozanov@gmail.com
#
# Copyright (C) 2018-Present Nikolai Rozanov
#
####################################################

####################################################
# IMPORT STATEMENTS
####################################################

# >>>>>>  Native Imports  <<<<<<<

# >>>>>>  Package Imports <<<<<<<
import numpy as np

# >>>>>>  Local Imports   <<<<<<<
from barl.utils import numerical

from .interfaces import BaseBayesianEstimator
####################################################
# CODE
####################################################


class NormalMean(BaseBayesianEstimator):
    """
    Estimating 1D Vectors
    """
    def __init__(self, size, initialVariance=3.0, samplingVariance=0.1):
        assert len(size) < 2, "EmpiricalMean only does 1D estimates at the moment"

        self.__means  = np.zeros(size)
        self.__variances = np.ones(size) * initialVariance
        self.__sampling_variances = np.ones(size) * samplingVariance

        self.__shape = self.__means.shape
        self.__ndim = self.__means.ndim

    def __str__(self):
        """
        Represents the empricial mean as a string
        """
        outStr = "Means: " + str(self.__means) + '\n' + "Variances: " + str(self.__variances)
        return outStr

    def update( self , newValueOrList, indexList=None):
        """
        updates the mean
        Can update from ListOf Values and List of Indexes

        """
        newValue = self.__deal_with_valueList( newValueOrList, indexList )
        oneHotIndexList = self.__deal_with_index( indexList )

        if newValue.shape == self.__shape:
            self.__update_single_value(newValue, oneHotIndexList)

        elif (newValue.ndim == self.__ndim+1) and (newValue.shape[1]==self.__shape[0]):
            self.__update_multiple_values(newValue, oneHotIndexList)




    def get_estimate(self, returnVariances=False):
        """
        returns the current estimate (of the mean)
        """
        if returnVariances:
            return self.__means, self.__variances

        else:
            return self.__means



    def sample(self):
        """
        returns the current estimate (of the mean)
        """
        # sampling the means
        samplingMeans = np.random.multivariate_normal( mean=self.__means, cov=np.diag(self.__variances) )

        # sampling the estimate
        sample = np.random.multivariate_normal( mean=samplingMeans, cov=np.diag(self.__sampling_variances) )

        return sample




    def __update_single_value(self, newValue, oneHotIndexList=None):
        """
        updates a single value
        """
        if type(oneHotIndexList)==type(None):
            where = np.ones(self.__shape)
        else:
            where = oneHotIndexList

        sampleMean = newValue

        self.__update_means(sampleMean, where)
        self.__update_variances(sampleMean, where)





    def __update_multiple_values(self, newValue, oneHotIndexList=None):
        """
        updates multiple values
        """
        if type(oneHotIndexList) == type(None):
            where =  np.ones(self.__shape) * newValue.shape[0]

        else:
            where = np.sum( oneHotIndexList, axis=0 )

        values = np.sum( newValue, axis=0 )
        sampleMean = np.divide(values, where, where=where>0)

        self.__update_means(sampleMean, where)
        self.__update_variances(sampleMean, where)




    def __deal_with_valueList(self, values, index=None):
        """
        creates valued one hot or does nothing
        """

        if type(index) == type(None):
            outValues = np.array(values)
        else:
            outValues = numerical.transform_to_valued_one_hot( listOfValues=values, listOfIndexes=index, numClasses=self.__shape[0] )

        return outValues




    def __deal_with_index(self,index=None):
        """
        creates a one shot or returns None
        """
        if type(index) == type(None):
            return None
        else:
            outIndex = numerical.transform_to_one_hot(index, numClasses=self.__shape[0])
            return outIndex



    def __update_means(self, sampleMean, countVector):
        """
        updates means according to Conjugate Prior update rule.

        https://people.eecs.berkeley.edu/~jordan/courses/260-spring10/lectures/lecture5.pdf

        (sg0^2 * x_bar + sg^2 * mu0) / (sg^2/n + sg0^2)

        sg == sampling variances
        sg0 == mean variances
        """

        mask = countVector>0

        sg2n = np.divide( self.__sampling_variances, countVector, where = mask)

        denominator = sg2n[mask] + self.__variances[mask]
        numerator = self.__variances[mask] * sampleMean[mask] + self.__sampling_variances[mask] * self.__means[mask]

        self.__means[mask] = numerator / denominator

    def __update_variances(self, sampleMean, countVector):
        """
        updates means according to Conjugate Prior update rule.

        https://people.eecs.berkeley.edu/~jordan/courses/260-spring10/lectures/lecture5.pdf

        1 / (1/sg0^2 + n/sg^2)

        sg == sampling variances
        sg0 == mean variances

        Note: sampleMean is not used
        """

        mask = countVector>0

        sg2n = np.divide( countVector, self.__sampling_variances, where = mask )
        sg0  = np.divide(1, self.__variances)

        denominator = sg2n[mask] + sg0[mask]

        self.__variances[mask] = 1 / denominator




####################################################
# MAIN
####################################################


# EOF



####################################################
#
# This Work is written by Nikolai Rozanov <nikolai>
#
# Contact:  nikolai.rozanov@gmail.com
#
# Copyright (C) 2018-Present Nikolai Rozanov
#
####################################################

####################################################
# IMPORT STATEMENTS
####################################################

# >>>>>>  Native Imports  <<<<<<<
import warnings
warnings.filterwarnings("ignore")

# >>>>>>  Package Imports <<<<<<<
import seaborn as sns
import matplotlib.pyplot as plt

# >>>>>>  Local Imports   <<<<<<<
from barl.utils.utils import unzip



####################################################
# CODE
####################################################
sns.set(style="whitegrid")

def action_reward_barplot(arList, show=True):
    """
    plots a barplot
    """
    data = transform_arlist_to_data(arList)

    sns.barplot(x="actions", y="rewards", data=data)

    if show:
        plt.show()


def plot_reward_over_time_from_ar(arList, show=True):
    """
    plots reward over time
    """
    data = transform_arlist_to_data(arList)

    sns.lineplot( x=data["time"], y=data["rewards"] )

    if show:
        plt.show()



def plot_reward_over_time(trList, show=True):
    """
    plots reward over time
    """
    data = transform_trlist_to_data(trList)

    sns.lineplot( x=data["time"], y=data["rewards"] )

    if show:
        plt.show()

def plot_actions_over_time_from_ar(arList, show=True):
    """
    plots reward over time
    """
    data = transform_arlist_to_data(arList)

    sns.lineplot( x=data["time"], y=data["actions"] )

    if show:
        plt.show()









def transform_trlist_to_data(trList):
    """
    """
    timeSteps, rewards = unzip(trList)

    data = {"rewards":list(rewards), "time": timeSteps}

    return data

def transform_arlist_to_data(arList):
    """
    """
    actions, rewards = unzip(arList)

    timeSteps = range( len(actions) )

    data = {"actions": list(actions), "rewards":list(rewards), "time": timeSteps}

    return data

####################################################
# MAIN
####################################################


# EOF



####################################################
#
# This Work is written by Nikolai Rozanov <nikolai>
#
# Contact:  nikolai.rozanov@gmail.com
#
# Copyright (C) 2018-Present Nikolai Rozanov
#
####################################################

####################################################
# IMPORT STATEMENTS
####################################################

# >>>>>>  Native Imports  <<<<<<<

# >>>>>>  Package Imports <<<<<<<

# >>>>>>  Local Imports   <<<<<<<
from barl.agents.interfaces import BaseStateLessAgent
from barl.environments.interfaces import StateLessEnvironment


####################################################
# CODE
####################################################

def run_state_less_agent_and_env(agent, environment, N=10):
    """
    Runs the simulation of agent and environment for N steps

    returns totalReward, ActionRewardList, NumOfRuns
    """
    assert issubclass( type(agent), BaseStateLessAgent ) , "Wrong Agent"
    assert issubclass( type(environment), StateLessEnvironment ) , "Wrong Environment"

    ARList = []
    totalReward = 0

    for t in range(N):

        action = int( agent.take_action() )
        reward = float( environment.sample_rewards(action=action) )

        ARList.append( (action, reward) )
        totalReward += reward

    return totalReward, ARList, N


def run_and_train_state_less_agent_and_env(agent, environment, N=10):
    """
    Runs the simulation of agent and environment for N steps, while training the agent

    returns totalReward, ActionRewardList, NumOfRuns
    """
    assert issubclass( type(agent), BaseStateLessAgent ) , "Wrong Agent"
    assert issubclass( type(environment), StateLessEnvironment ) , "Wrong Environment"

    ARList = []
    totalReward = 0

    for t in range(N):

        action = int( agent.take_action() )
        reward = float( environment.sample_rewards(action=action) )

        arTuple = ( action, reward )
        agent.learn( [arTuple] )

        ARList.append( arTuple )
        totalReward += reward

    return totalReward, ARList, N








####################################################
# MAIN
####################################################


# EOF



####################################################
#
# This Work is written by Nikolai Rozanov <nikolai>
#
# Contact:  nikolai.rozanov@gmail.com
#
# Copyright (C) 2018-Present Nikolai Rozanov
#
####################################################

####################################################
# IMPORT STATEMENTS
####################################################

# >>>>>>  Native Imports  <<<<<<<

# >>>>>>  Package Imports <<<<<<<
import numpy as np

# >>>>>>  Local Imports   <<<<<<<
from barl.utils import utils
from .interfaces import BaseStateLessAgent

####################################################
# CODE
####################################################

class RandomActionsSampler(BaseStateLessAgent):
    """
    simply random actions
    """
    def __init__(self,numActions):
        super().__init__(numActions)

    def learn(self, arList):
        pass

    def take_action(self):
        action =  np.random.choice( self.numActions, 1 )

        return np.squeeze(action)



class FixedActionsSampler(BaseStateLessAgent):
    """
    simply random actions
    """
    def __init__(self,numActions, fixedAction):
        super().__init__(numActions)
        self.fixedAction = fixedAction

    def learn(self, arList):
        pass

    def take_action(self, action=None):
        if type(action)==type(None):
            action = self.fixedAction

        return action







####################################################
# MAIN
####################################################


# EOF



####################################################
#
# This Work is written by Nikolai Rozanov <nikolai>
#
# Contact:  nikolai.rozanov@gmail.com
#
# Copyright (C) 2018-Present Nikolai Rozanov
#
####################################################

####################################################
# IMPORT STATEMENTS
####################################################

# >>>>>>  Native Imports  <<<<<<<
from setuptools import setup, find_packages

# >>>>>>  Package Imports <<<<<<<

# >>>>>>  Local Imports   <<<<<<<


####################################################
# CODE
####################################################


with open("README.md", "r") as fh:
    longDescription = fh.read()


####################################################
# MAIN
####################################################
setup(

      #
      # SETUP
      #

      name          ='barl',
      version       ='0.0.0.5',
      url           ='https://github.com/ai-nikolai/barl',
      author        ='nikolai rozanov',
      author_email  ='nikolai.rozanov@gmail.com',
      license       ='Apache 2.0',

      #
      # Actual packages, data and scripts
      #

      packages      = find_packages(),

      package_dir   = {'barl': 'barl'},

      install_requires=[
                        "numpy",
                        "matplotlib",
                        "seaborn>=0.9",
                       ],

      python_requires='>=3.6',

      #
      # DESCRIPTION
      #
      description   ='Bayesian Approximate Reinforcement Learning (BARL)',
      long_description=longDescription,
      long_description_content_type="text/markdown",
    )
# EOF



####################################################
#
# This Work is written by Nikolai Rozanov <nikolai>
#
# Contact:  nikolai.rozanov@gmail.com
#
# Copyright (C) 2018-Present Nikolai Rozanov
#
####################################################

####################################################
# IMPORT STATEMENTS
####################################################

# >>>>>>  Native Imports  <<<<<<<
import os
import sys
from functools import partial


# >>>>>>  Package Imports <<<<<<<
import numpy as np

# >>>>>>  Local Imports   <<<<<<<
from barl import BarlException
from .interfaces import StateLessEnvironment


####################################################
# CODE
####################################################

class MultiArmedBandit(StateLessEnvironment):
    """
    MultiArmedbandit - MAB
    """
    #Allowed Distributions
    distributionMapping = {
        "Normal" : np.random.normal,
    }


    def __init__(self, arms=None, means=None, variances=None, distribution="Normal"):
        if arms:
            self.__arms = arms
            if not means:
                means = MultiArmedBandit.get_means( arms )

            if not variances:
                variances = MultiArmedBandit.get_variances( arms )
        else:
            if not means:
                raise BarlException("Need to provide Means to Multiarmed Bandit")
            else:
                self.__arms = len(means)

            if not variances:
                variances = MultiArmedBandit.get_variances( self.__arms )

        self.__means = means
        #to allow single number input
        flag = type(variances)==np.ndarray or type(variances)==list
        self.__variances = variances if flag else np.ones(self.__arms) * variances

        self.__distribution = distribution

        self.__set_arms()


    def sample_rewards(self, action=None, N=1):
        """
        samples rewards: np.array ( size[arms,N] )
        """
        if type(action)==type(None):

            outList = []

            for arm in self.arms:

                outList.append( arm(N) )

            rewards = np.array( outList )

        else:

            rewards = self.arms[ int(action) ](N)

        return np.squeeze( rewards )



    def get_true_variances(self):
        """
        """
        return self.__variances

    def get_true_means(self):
        """
        """
        return self.__means


    def __set_arms(self):
        """
        sets the arm properties
        """

        self.arms = []

        distro = MultiArmedBandit.distributionMapping[self.__distribution]

        for mean, variance in zip(self.__means,self.__variances):

            temp = partial(distro, mean, variance)

            self.arms.append(temp)


    @staticmethod
    def get_variances(arms = 2):
        """
        returns [v1,v2, ...v_arms] sampled from ~Unif(0,1)
        """
        return np.random.beta(a=1,b=1,size=[arms])


    @staticmethod
    def get_means(arms = 2):
        """
        returns [m1,m2, ...m_arms] sampled from ~Normal(0,1)
        """
        return np.random.normal(loc=0.0, scale=1.0, size=[arms])




####################################################
# MAIN
####################################################


# EOF
